;=============[ LCDCMD.ASM ]==========================[ 12-14-2002 ]==;
; (C) 2002 Jerry Han
; Protected under the GNU General Public License
; for specific license information, see www.gnu.org/copyleft/gpl.html
;
; version 1.00 PRERELEASE
;
; This piece of software is designed to run on a PIC (16F628 or equivalent)
; and connect to a standard (Hitachi 44780-based) LCD/VFD or similar display.
;
; It provides an RS-232 interface on which it received commands for
; the display.  It attempts to emulate the Matrix-Orbital serial display
; command set.
;
; ============================================================================
; This file contain some low level LCD control routine
;
;-------------LCD comm & control routines--------------------------

;-------------delay routines---------------------------------------

;Time Delay = .0005012 s with Osc = 10.000000 MHz
DELAY500
	if _ClkIn == 20000000
		movlw	D'4'			; 500usec delay for 20M osc
	else
		movlw	D'2'			; 500usec delay for 10M or 4M osc
	endif
		movwf	ODELAY
	if _ClkIn >= 10000000
		movlw	D'158'			; for 10M or 20M osc
	else
		movlw	D'63'			; for 4M osc
	endif
		movwf	DELAY
D500LOOP	decfsz	DELAY,1
		goto	D500LOOP
		decfsz	ODELAY,1
		goto	D500LOOP
D500END		return

XDELAY500	movwf	X_DELAY
XDELAY500LOOP	call	DELAY500
		decfsz	X_DELAY,f
		goto	XDELAY500LOOP
XDELAY500END	return



;-------------LCD initialization-----------------------------------


LCD_INIT	call	LCD_INITDLY
						; okay now initialize the LCD and select 4-bit interface
		clrf	LCD_DATA
		bsf	LCD_DATA,D4
		bsf	LCD_DATA,D5
		call	PULSE_E_NOBF
		movwf	d'10'			; wait at least 4.1ms (5ms here)
		call	XDELAY500
		call	PULSE_E_NOBF
		call	PULSE_E_NOBF		; initialization done.

		bcf	LCD_DATA,D4
		call	PULSE_E_NOBF		; 4-bit interface selected.
		call	DELAY500
						; now we can use the busy-flag!
		movlw	b'00101100'		; function set (2-line, 5x7 font)
		call	LCD_CMD
		movlw	b'00001100'		; display command (display on, no cursor, no blink)
		call	LCD_CMD
		call	LCD_INITDLY
		return

LCD_INITDLY					;delay = 1.0000012 s with Osc = 10.000000 MHz
		movlw	D'5'			; delay = 0.5s
		movwf	BUFFERB
LCD_INITLP	movlw	D'200'			; delay = 0.1s
		call	XDELAY500
		decfsz	BUFFERB,f
		goto	LCD_INITLP
		return

;-------------LCD put-character------------------------------------

LCD_PUTCHAR	call	LCD_BUSY			; make sure device is ready!
		bsf	LCD_CTL,LCD_RS			; raise RS
		call	LCD_SEND			; send the character
		incf	D_ADDR,f			; need to figger out what to do with d_addr at wraptime
		return

;-------------LCD put cg-character------------------------------------

LCD_CGPUTCHAR	call	LCD_BUSY			; make sure device is ready!
		bsf	LCD_CTL,LCD_RS			; raise RS
		call	LCD_SEND			; send the character
		incf	CG_ADDR,f			; need to figger out what to do with d_addr at wraptime
		return


;-------------LCD send a byte--------------------------------------
; (note that this routine makes assumptions
;  regarding Pin placement for LCD_DATA)
LCD_SEND	movwf	BUFCHAR				; save the character
		movwf	BUFFERA				; now select only the top 4 bits
		swapf	BUFFERA,w
		andlw	0x0f
		movwf	LCD_DATA			; push them into the data register
		bsf	LCD_CTL,LCD_E			; send them to the LCD
		nop
		nop
		nop
		bcf	LCD_CTL,LCD_E
		movf	BUFCHAR,w			; reload the original character
		movwf	LCD_DATA			; push the lower 4 bits into place
		bsf	LCD_CTL,LCD_E			; and send them to the LCD
		nop
		nop
		nop
		bcf	LCD_CTL,LCD_E
		bcf	LCD_CTL,LCD_RS
		return					; note that this returns with W intact

LCD_GETCHAR
		call	LCD_BUSY			; make sure device is ready!


		bsf	LCD_CTL,LCD_RS			; set data to be read
		bsf	LCD_CTL,LCD_RW
		nop
		bsf	LCD_CTL,LCD_E			; require the data
		nop
		bsf	STATUS, RP0			; set to bank1
		movlw	0x0f				; set the port for input
		movwf	TRISA
		bcf	STATUS, RP0

		movf	LCD_DATA, w			; get first bibble data
		bcf	LCD_CTL,LCD_E
		andlw	0x0f
		movwf	BUFFERA
		swapf	BUFFERA,f

		bsf	LCD_CTL,LCD_E			; require the data
		nop
		movf	LCD_DATA, w			; get next bibble data
		andlw	0x0f
		iorwf	BUFFERA,w			; Combine the 8bit data
		bcf	LCD_CTL,LCD_E
		bcf	LCD_CTL,LCD_RS
		bcf	LCD_CTL,LCD_RW

		bsf	STATUS, RP0			; set to bank1
		clrf	TRISA
		bcf	STATUS, RP0

		return

;-------------LCD execute command----------------------------------

LCD_CMD		call 	LCD_BUSY			; make sure the device is ready!
		bcf	LCD_CTL,LCD_RS			; lower RS
		call	LCD_SEND			; send the command
		return

;-------------LCD busy-flag checker!-------------------------------
LCD_BUSY	movwf	BUFFERW				; save the contents of W
		movf	LCD_DATA,w
		movwf	BUFFERA				; save the contents of LCD_DATA
		movf	LCD_CTL,w
		movwf	BUFFERB				; save the contents of LCD_CTL
		clrf	LCD_DATA
		bsf	STATUS,RP0
		movlw	0x0f
		movwf	TRISA				; change LCD_DATA to inputs!
		bcf	STATUS,RP0
		bcf	LCD_CTL,LCD_RS			; set up the busy-flag request pin config
		bsf	LCD_CTL,LCD_RW			; set up the busy-flag request pin config
BUSYLOOP	bsf	LCD_CTL,LCD_E			; pulse E to get the busy flag
		nop
		nop
		nop
		movf	LCD_DATA,w			; load the busy flag status into W
		bcf	LCD_CTL,LCD_E			; release E.
		nop
		nop
		nop
		bsf	LCD_CTL,LCD_E			; have to pulse E again (to complete the BF
		nop					; sequence). we don't care about d_addr.
		nop
		nop
		bcf	LCD_CTL,LCD_E			; turn off E an
		andlw	BF_FLAG				; and check W for busy flag
		btfss	STATUS,Z			; is LCD available?
		goto	BUSYLOOP			; no it isn't
LCD_NOTBUSY	bsf	STATUS,RP0			; yes it is
		clrf	TRISA				; LCD_DATA is all outputs again!
		bcf	STATUS,RP0
		movf	BUFFERA,w
		movwf	LCD_DATA			; restore the contents of LCD_DATA
		movf	BUFFERB,w
		movwf	LCD_CTL				; restore the contents of LCD_CTL
		movf	BUFFERW,w			; restore the contents of W
		return


;-------------pulse the E line without BF check--------------------

PULSE_E_NOBF	call	DELAY500			; no busy-flag check - use a built-in delay instead
		call	DELAY500
		bsf	LCD_CTL,LCD_E
		nop
		nop
		nop
		bcf	LCD_CTL,LCD_E
		return

;=============================================================================
; LEDON
; Sends command to turn on the LED
; Required command must be in W
; OK
;=============================================================================
LEDON
		movlw	TYPE_VFD
		xorwf	DISPLAY_TYPE,w
		btfss	STATUS,Z
		goto	_LEDON
		bsf	FLAGS, VFD_E		; turn on VFD
		call	LCD_CONTROL
		return
_LEDON		BSF	LCD_CTL, LED;
		RETURN

;=============================================================================
; LEDOFF
; Sends command to turn off the LED
; OK
;=============================================================================
LEDOFF
		movlw	TYPE_VFD
		xorwf	DISPLAY_TYPE,w
		btfss	STATUS,Z
		goto	_LEDOFF
		bcf	FLAGS, VFD_E		; turn off VFD
		call	LCD_CONTROL
		return
_LEDOFF		BCF	LCD_CTL, LED;
		RETURN


LCD_CONTROL
		movlw	b'00000111'
		andwf	FLAGS,w			; load the BLINK and CURSOR flags into W
		iorlw	b'00001000'		; create a display control command
		call	LCD_CMD
		return

;-------------splashscreen-----------------------------------------

ShowSplash
		clrf	POS_X
		clrf	POS_Y

		movlw	low EE_SPLASH
		movwf	TABLE_INDEX

NewPos		call	GET_ADDRESS
		call	MOVE_CURSOR

ProcNext	movf	TABLE_INDEX,w
		call	EE_READ
		call	LCD_PUTCHAR
		incf	TABLE_INDEX,f

		incf	POS_X,f
		movf	DISPLAY_WIDTH,w
		xorwf	POS_X,w
		btfss	STATUS,Z		; end of line?
		goto	ProcNext		; no, process next
		clrf	POS_X			; wrap to start of next line
		incf	POS_Y,f
		movf	DISPLAY_LINES,w
		xorwf	POS_Y,w
		btfss	STATUS,Z		; last line?
		goto	NewPos			; no, set the D_ADDR from new pos
		clrf	POS_Y			; yes, exit the loop
		return

SPLASHSCREEN
		goto	ShowSplash		; no delay, and leave the init screen on LCD
		call	ShowSplash

ShowEnd		movlw	0x15
		movwf	TABLE_INDEX
MYLOOP		movlw	0xff			; wait for a while so the user can appreciate
		call	XDELAY500		; the glory and splendor of the splashscreen
		DECFSZ	TABLE_INDEX,f
		goto	MYLOOP
SPLASHEND	movlw	LCDCLEAR		; clear the screen
		call	LCD_CMD
		clrf	D_ADDR			; keep our memory pointer accurate
		return


; read a byte from EEPROM
; the address is in W reg, the result is in W reg
EE_READ
		bsf	STATUS, RP0
		movwf	EEADR
		bsf	EECON1, RD
		movf	EEDATA, w
		bcf	STATUS, RP0
		return

EE_WRITE
		; set up the data and the address
		bsf	STATUS, RP0 		; Bank 1

		btfsc	EECON1,WR
		goto	$-1

		bcf	STATUS, RP0 		; Bank 0
		movf	CONFIG_ADDR,w		;
		bsf	STATUS, RP0 		; Bank 1
		movwf	EEADR			; Address to write
		bcf	STATUS, RP0 		; Bank 0
		movf	CONFIG_DATA,w		;
		bsf	STATUS, RP0 		; Bank 1
		movwf	EEDATA			; Data to write
						; perform the write operation
		bsf	EECON1, WREN		; Enable Write
		bcf	INTCON, GIE		; Disable INTs
		movlw	055h			;
		movwf	EECON2			; Write 55
		movlw	0AAh			;
		movwf	EECON2			; Write AA
		bsf	EECON1, WR		; Set WR bit

		btfsc	EECON1,WR
		goto	$-1

		bsf	INTCON, GIE		; Enable the INTs
		bcf	EECON1,WREN
		bcf	STATUS, RP0		; Bank 0
		return
