;=============[ MO-LCD.ASM ]==========================[ 12-14-2002 ]==;
; (C) 2002 Jerry Han
; Protected under the GNU General Public License
; for specific license information, see www.gnu.org/copyleft/gpl.html
;
; version 1.00 RELEASE
;
; This piece of software is designed to run on a PIC (16F628 or equivalent)
; and connect to a standard (Hitachi 44780-based) LCD/VFD or similar display.
;
; It provides an RS-232 interface on which it received commands for
; the display.  It attempts to emulate the Matrix-Orbital serial display
; command set.
;
; ============================================================================
; WATCH OUT -
;
; Not all M-O commands are implemented - see below.
;
;
; if OSC = 10M
; 1 cycle time = 0.4us = 400ns

	LIST P=16F628, F=INHX8M
	__CONFIG _CP_OFF & _WDT_OFF & _HS_OSC & _LVP_OFF & _BODEN_OFF & _DATA_CP_OFF & _PWRTE_ON

	include <p16f628.inc>

;=============EQUATES==============================================
TRUE		equ	1
FALSE		equ	0

USE_EEPROM	equ	1

;*********************************************************************************************************
;                               Setup RS-232 Parameters
;*********************************************************************************************************

;_ClkIn		equ	4000000         ; Input Clock Frequency is 4 Mhz
_ClkIn		equ	20000000         ; Input Clock Frequency is 20 Mhz
;_ClkIn		equ	10000000         ; Input Clock Frequency is 10 Mhz
_BaudRate       set     19200            ; Baud Rate (bits per second) is 1200
_DataBits       set     8               ; 8 bit data, can be 1 to 8
_StopBits       set     1               ; 1 Stop Bit, 2 Stop Bits is not implemented

;-------------equates for pins and ports---------------------------

LCD_DATA	equ	PORTA
LCD_CTL		equ	PORTB
D4		equ	0x00		; PORTA,pin0 = 44780,DB4 (pin11)
D5		equ	0x01		; PORTA,pin1 = 44780,DB5 (pin12)
D6		equ	0x02		; PORTA,pin2 = 44780,DB6 (pin13)
D7		equ	0x03		; PORTA,pin3 = 44780,DB7 (pin14)
BF_FLAG		equ	b'00001000'	; should have a 1 where D7 is on LCD_DATA

LCD_E		equ	0x03		; PORTB,pin3 = 44780,E   (pin6)
LCD_RS		equ	0x00		; PORTB,pin0 = 44780,RS  (pin4)
LCD_RW		equ	0x05		; PORTB,pin5 = 44780,R/W (pin5)
LED		equ	0x04		; PORTB,pin4 = Backlight LED
GPO1		equ	0x06		; PORTB,pin6 = GPO1
GPO2		equ	0x07		; PORTB,pin7 = GPO2

TYPE_VFD	equ	0		; type define for VFD module
TYPE_LCD	equ	1		; type define for LCD module


;-------------equates for variables--------------------------------
	CBLOCK	0x20
		DELAY			; main delay variable
		ODELAY			; main delay outer-loop variable
		X_DELAY			; counter for a multi-delay loop
		FLAGS			; state of serial display
		BUFFERA			; place to store PORTA settings during ready check
		BUFFERB			; place to store PORTB settings during ready check
		BUFFERW			; place to store W settings during ready check
		BUFCHAR			; character being sent to LCD
		D_ADDR			; memory location of cursor in LCD
		POS_X			; cursor position X(char number)
		POS_Y			; cursor position Y(line number)
		COMFLAG			; communications flag
		RECREG			; received data character
		CG_ADDR			; character generator address
		TABLE_INDEX		; Index to table strings
		R_ADDR			; DDRAM address for read
		W_ADDR			; DDRAM address for write
		TMPW			; temporary W
		SaveWReg
		SaveStatus
		ITMPW
		TEMPW
		DTMPW
		CurrentLine
		G_CHAR
		PWM_PERIOD
		PWM_DUTY
		CONFIG_ADDR
		CONFIG_DATA
		DISPLAY_TYPE		; Display Type
		DISPLAY_WIDTH		; Display width, default 20
		DISPLAY_LINES		; Display line number, default 2
		RECVBUF_COUNT		; the char numbers in receive buffer
		RECVBUF_READ_IDX	; the read index of receive buffer
	ENDC

RECVBUFF_ADDR	equ	0xA0		; first byte of receive buffer, on bank 2
RECVBUFF_SIZE	equ	0x100 - RECVBUFF_ADDR	; size of receive buffer, 96 bytes here. :)
RECVBUFF_END	equ	0xff		; End address of receive buffer


;-------------bit equates for flag variable------------------------

BLINK		equ	0x00		; Blink Cursor
CURSOR		equ	0x01		; Show Cursor
VFD_E		equ	0x02		; Enable VFD (should always be 1 as LCD)
LINEWRAP	equ	0x04		; Internal Line Wrap flag
AUTOSCROLL	equ	0x05		; Internal Auto Scroll up flag
EE_BUSY		equ	0x06		; EEPROM writting busy flag

;-------------equates for LCD commands & LCD mem locations-------------
LCDCLEAR	equ	b'00000001'
LCD_LINE0	EQU	0x000		; ddram location of line 0
LCD_LINE1	EQU	0x040		; ddram location of line 1
LCD_LINE2	EQU	0x014		; ddram location of line 2
LCD_LINE3	EQU	0x054		; ddram location of line 3
LCD_WIDTH	EQU	d'20'		; Default How many chars in one line
LCD_LINES	EQU	d'2'		; Default How many lines
MTRXCMD		equ	0xfe		; Matrix Orbital Command Character

;=============BEGIN PROGRAM========================================

		org	0x00
		goto	START

		org	0x04			; interrupt service routine
		goto	ISR

		include "tables.asm"		; The lookup table routines are in file "tables.asm"

START
		bcf	STATUS,RP0
		clrf	PORTA			; PORTA all low
		clrf	PORTB			; PORTB all low
		movlw	0x07			; turn comparators off and
		movwf	CMCON			; enable pins for I/O functions
		bcf	STATUS,RP1
		bsf	STATUS,RP0		; switch to bank 1
		clrf	TRISA			; PORTA all outputs
		movlw	b'00000110'
		movwf	TRISB
		bcf	STATUS,RP0		; back to bank 0
		call	InitUSART		; Initialize USART
		clrf	D_ADDR			; clear the daddr pointer.
		clrf	POS_X
		clrf	POS_Y
		clrf	FLAGS			; clear all the flags
		bsf	FLAGS,LINEWRAP		; start with linewrap enabled.  (this might be wrong!)
		bsf	FLAGS,VFD_E		; enable the display flag
		call	Initialize		; Initialize system setting from EEPROM
		call	LCD_INIT		; Initialize the LCD module

		movlw	LCDCLEAR		; clear the screen
		call	LCD_CMD

		call	LEDON			; turn on the backlight
		call	SPLASHSCREEN		; put an advertisement or splashscreen in right here!

		;call	SHOW_INFO

		;goto 	CURSORON		; turn cursor, better for debug

						; begin main program loop
PROCLOOP
		call	GETCHAR			; won't return until a character has been received correctly
		movlw	MTRXCMD			; (character is stored in RECREG)
		xorwf	RECREG,w
		btfsc	STATUS,Z		; is it the Matrix Orbital command character?
		goto	GETCOMMAND		; yes it's a command
						; no its data so fall through to here

		movlw	0x0c			; char FF, clear the display and put the cursor at the top left
		xorwf	RECREG,w
		btfsc	STATUS,Z		; is a ClearScreen command
		goto	CLEARDISP		; clear display

		movlw	0x08			; char Backspace
		xorwf	RECREG,w
		btfsc	STATUS,Z		; Is a BackSpace char
		goto	BACKSPACE		; move cursor left one char

		movlw	0x0a			; char Line Feed, move the cursor to beginning of next line
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	DO_NEWLINE		; Scroll one line up

		movlw	0x0d			; char CR, move the cursor to beginning of current line
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	DO_RETURN

WRITEDATA	movf	RECREG,w		; load the character
		call	LCD_PUTCHAR		; and display it
		call	CUR_RIGHT
		goto	PROCLOOP		; return to main loop

;--------------matrix-orbital command processing--------------------

GETCOMMAND	call	GETCHAR			; have to get the next character in the command sequence
		movlw	0x3d			; CMD: Make Vertical Bar Graph
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	MAKE_VERT_BAR

		movlw	0x7c			; CMD: Make Horizontal Bar Graph
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	MAKE_HORI_BAR

		movlw	0x68			; CMD: Initialize Horizontal Bar Graph
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	INIT_HORI_BAR

		movlw	0x73			; CMD: Initialize Narrow Vertical Bar Graph
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	INIT_VERT_NBAR

		movlw	0x76			; CMD: Initialize Wide Vertical Bar Graph
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	INIT_VERT_WBAR

		movlw	'5'			; CMD: Read Serial No
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	PROCLOOP

		movlw	'6'			; CMD: Read Version
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	PROCLOOP

		movlw	'7'			; CMD: Read Module Type
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	PROCLOOP

		movlw	'&'			; CMD: Pool Keypad
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	PROCLOOP

		movlw	'n'			; CMD: ???
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	PROCLOOP

		movlw	'#'			; CMD: ???
		xorwf	RECREG,w
		btfsc	STATUS,Z
		goto	IGN_2

		movlw	HIGH JUMP_TABLE
		movwf	PCLATH			; TODO: see if you have to undo this somehow....
		movlw	b'00011111'
		andwf	RECREG,w		; grab the lowest 5 bytes of the character  (0 through 31)
		addwf	PCL,f			; jump by alphabetic numbers
JUMP_TABLE	goto	PROCLOOP		;@
		goto	PROCLOOP		;A
		goto	BACKLIGHTON		;B
		goto	LINEWRAPON		;C
		goto 	LINEWRAPOFF		;D
		goto	CLEARKEYBUF		;E
		goto 	BACKLIGHTOFF		;F
		goto	GOTOPOS			;G
		goto	CURSORHOME		;H (command = top left)
		goto	PROCLOOP		;I - not used
		goto	CURSORON		;J
		goto	CURSOROFF		;K
		goto	CURSORLEFT		;L
		goto	CURSORRIGHT		;M
		goto	CUSTOMCHAR		;N
		goto 	PROCLOOP		;O -not used
		goto 	CONTRAST		;P set contrast
		goto 	SCROLLON		;Q
		goto 	SCROLLOFF		;R
		goto	BLINKON			;S
		goto	BLINKOFF		;T
		goto	DEBOUNCE		;U
		goto	GPOFF			;W
		goto	GPON			;V
		goto	CLEARDISP		;X
		goto	BRIGHTNESS		;Y
		goto	SETTYPE			;Z
		goto	LOADSPLASH		;[
		goto	PROCLOOP		;\
		goto	PROCLOOP		;]
		goto	PROCLOOP		;^
		goto	PROCLOOP		;_
JUMP_TABLE_END	goto	PROCLOOP		;` auto repeat mode off


	IF ( (JUMP_TABLE & 0x0FF) >= (JUMP_TABLE_END & 0x0FF) )
		MESSG   "Warning - User Definded: Table 'JUMP_TABLE' crosses page boundry in computed jump"
	ENDIF

IGN_2
		call	GETCHAR
IGN_1
		call	GETCHAR
IGN
		goto	PROCLOOP
		
LINEWRAPON	bsf	FLAGS,LINEWRAP
		goto 	PROCLOOP

LINEWRAPOFF	bcf	FLAGS,LINEWRAP
		goto	PROCLOOP

CURSORHOME
		clrf	POS_X
		clrf	POS_Y
		call	GET_ADDRESS
		call	MOVE_CURSOR
		goto	PROCLOOP

CURSORLEFT
		call	CUR_LEFT
		goto	PROCLOOP

CURSORRIGHT	call	CUR_RIGHT
		goto	PROCLOOP

LOADSPLASH	call	SaveSplash
		call	ShowSplash
		goto	PROCLOOP

SETTYPE
		movlw	.200
		call	XDELAY500		; delay 50ms, let command been transfered completely.

		movlw	low EE_WIDTH
		movwf	CONFIG_ADDR

		call	GETCHAR			; get width
		movf	RECREG,w
		movwf	DISPLAY_WIDTH

		call	GETCHAR			; get lines
		movf	RECREG,w
		movwf	DISPLAY_LINES

		call	GETCHAR			; get type
		movf	RECREG,w
		movwf	DISPLAY_TYPE

		movf	DISPLAY_WIDTH,w
		movwf	CONFIG_DATA
		call	EE_WRITE

		movf	DISPLAY_LINES,w
		movwf	CONFIG_DATA
		incf	CONFIG_ADDR,f
		call	EE_WRITE

		movf	DISPLAY_TYPE,w
		movwf	CONFIG_DATA
		incf	CONFIG_ADDR,f
		call	EE_WRITE
		goto	PROCLOOP

BLINKON		bsf	FLAGS,BLINK
		goto	DISPLAYCONTROL

BLINKOFF	bcf	FLAGS,BLINK
		goto	DISPLAYCONTROL

CURSORON	bsf	FLAGS,CURSOR
		goto	DISPLAYCONTROL

CURSOROFF	bcf	FLAGS,CURSOR
						; fall through to display control

DISPLAYCONTROL	movlw	b'00000111'
		andwf	FLAGS,w			; load the BLINK and CURSOR flags into W
		iorlw	b'00001000'		; create a display control command
		call	LCD_CMD
		goto 	PROCLOOP

SCROLLON	bsf	FLAGS,AUTOSCROLL	; turn on auto scroll
		goto	PROCLOOP

SCROLLOFF	bcf	FLAGS,AUTOSCROLL	; turn off auto scroll
		goto	PROCLOOP

BACKLIGHTON	call	GETCHAR			; number of minutes to keep backlight on
		call	LEDON
		goto	PROCLOOP		; command not implemented

BACKLIGHTOFF	call	LEDOFF
		goto	PROCLOOP		; commands not implemented

CONTRAST	call	GETCHAR			; get contrast value
		movlw	TYPE_LCD
		xorwf	DISPLAY_TYPE,w
		btfsc	STATUS,Z
		call	SET_CONTRAST
		goto	PROCLOOP

GPON		call	GETCHAR			; get output port
		call	GPO_ON
		goto	PROCLOOP		; commands not implemented

GPOFF		call	GETCHAR			; get output port
		call	GPO_OFF
		goto	PROCLOOP		; commands not implemented

CLEARKEYBUF	goto	PROCLOOP		; commands not implemented

CUSTOMCHAR	call	GETCHAR			; customer char #  (0=7)
		movf	RECREG,w
		andlw	b'00000111'		; only keep bits 0-2
		movwf	CG_ADDR
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f		; bits 3-5 now represent character number
		movf	CG_ADDR,w
		xorlw	b'01000000'		; load up a "set cgram address command"
		andlw	b'01111000'		; make sure only bits 3-6 are set at most
		call	LCD_CMD			; set CGRAM address
		call	GETCHAR			; customer char byte 1
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 2
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 3
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 4
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 5
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 6
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 7
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		call	GETCHAR			; customer char byte 8
		movf	RECREG,w
		call	LCD_CGPUTCHAR
		movf	D_ADDR,w
		iorlw	b'10000000'		; set bit 7 to form the set ddram address cmd
		call	LCD_CMD			; set ddram address
		goto	PROCLOOP
DEBOUNCE	call	GETCHAR			; debounce time
		goto	PROCLOOP		; command not implemented

INIT_HORI_BAR
		call	InitHorizontalBar
		goto	PROCLOOP
INIT_VERT_NBAR
		call	InitNarrowVerticalBar
		goto	PROCLOOP
INIT_VERT_WBAR
		call	InitWideVerticalBar
		goto	PROCLOOP
MAKE_VERT_BAR
		call	MakeVerticalBar
		goto	PROCLOOP
MAKE_HORI_BAR
		call	GETCHAR
		call	GETCHAR
		call	GETCHAR
		call	GETCHAR
		goto	PROCLOOP

GOTOPOS		call	GETCHAR			; fetch the column
		decf	RECREG,w		; subtract 1 from the entry
		;-------------------------------------
		; TODO: SHOULD VALID THE COLUMN VALUE
		;-------------------------------------
		movwf	POS_X			; set d_addr equal to column #
		call	GETCHAR			; fetch the row
		movf	RECREG,w
		subwf	DISPLAY_LINES,w
		btfss	STATUS,C
		goto	RowError
GOTOPOS1	decf	RECREG,w
		movwf	POS_Y

GOTOPOS_END
		call	GET_ADDRESS
		call	MOVE_CURSOR
		goto	PROCLOOP

RowError	movf	DISPLAY_LINES,w
		movwf	RECREG
		goto	GOTOPOS1

CLEARDISP	movlw	LCDCLEAR
		call	LCD_CMD			; clear the display
		clrf	D_ADDR			; keep our memory pointer accurate
		clrf	POS_X
		clrf	POS_Y
		goto	PROCLOOP

BRIGHTNESS	call	GETCHAR			; brightness level
		movlw	b'00000011'
		andwf	RECREG, w
		iorlw	b'00101100'		; function set (2-line, 5x7 font)
		call	LCD_CMD
		goto	PROCLOOP

BACKSPACE
		call	CUR_LEFT
		movlw	' '
		call	LCD_PUTCHAR		; Fill space on left position
		call	GET_ADDRESS
		call	MOVE_CURSOR		; Move cursor to previos position
		goto	PROCLOOP

; process the 'RETURN' key
; move the cursor to the beginning of current line.
DO_RETURN
		clrf	POS_X
		call	GET_ADDRESS
		call	MOVE_CURSOR
		goto	PROCLOOP
DO_NEWLINE
		decf	DISPLAY_WIDTH,w
		movwf	POS_X
		call	CUR_RIGHT
		goto	PROCLOOP

;-------------------------------------------------------------------------
; Scroll up on line
;-------------------------------------------------------------------------
SCROLL_UP
		clrf	CurrentLine

NextLine
		movlw	1
		subwf	DISPLAY_LINES,w
		xorwf	CurrentLine,w
		btfsc	STATUS,Z
		goto	FILL_LAST

		movlw	0
		xorwf	CurrentLine,w
		btfsc	STATUS,Z
		goto	S_LINE1

		movlw	1
		xorwf	CurrentLine,w
		btfsc	STATUS,Z
		goto	S_LINE2

		movlw	2
		xorwf	CurrentLine,w
		btfsc	STATUS,Z
		goto	S_LINE3
		goto	FILL_LAST

S_LINE1
		movlw	LCD_LINE1
		movwf	R_ADDR
		movlw	LCD_LINE0
		movwf	W_ADDR
		goto	LINE_BEGIN

S_LINE2
		movlw	LCD_LINE2
		movwf	R_ADDR
		movlw	LCD_LINE1
		movwf	W_ADDR
		goto	LINE_BEGIN

S_LINE3
		movlw	LCD_LINE3
		movwf	R_ADDR
		movlw	LCD_LINE2
		movwf	W_ADDR

LINE_BEGIN	movf	DISPLAY_WIDTH,w
		movwf	TABLE_INDEX

NEXT_CHAR
		movlw	0x80
		iorwf	R_ADDR,w
		call	LCD_CMD
		call	LCD_GETCHAR		; get the char in org line
		movwf	TMPW

		movlw	0x80
		iorwf	W_ADDR,w
		call	LCD_CMD
		movf	TMPW,w
		call 	LCD_PUTCHAR		; put the char in line 1

		incf	R_ADDR, f
		incf	W_ADDR, f
		DECFSZ	TABLE_INDEX, f
		goto	NEXT_CHAR

		incf	CurrentLine,f
		goto	NextLine

FILL_LAST	movf	DISPLAY_WIDTH,w
		subwf	R_ADDR,f
		movf	DISPLAY_WIDTH,w
		movwf	TABLE_INDEX
		movlw	0x80
		iorwf	R_ADDR,w
		call	LCD_CMD
		movf	R_ADDR,w
		movwf	D_ADDR

NEXT_CHAR2	movlw	0x20
		call 	LCD_PUTCHAR		; file space at line 1
		DECFSZ	TABLE_INDEX, f
		goto	NEXT_CHAR2

		return

;-------------------------------------------------------------------------
; Turn on the GPO output
;-------------------------------------------------------------------------
GPO_ON		movlw	0x01
		xorwf	RECREG,w		; Test for GPO 1
		btfsc	STATUS, Z
		bsf	PORTB,GPO1		; Enable the GPO 1 output
		movlw	0x02
		xorwf	RECREG,w		; Test for GPO 2
		btfsc	STATUS, Z
		bsf	PORTB,GPO2		; Enable the GPO 2 output
		return

;-------------------------------------------------------------------------
; Turn off the GPO output
;-------------------------------------------------------------------------
GPO_OFF		movlw	0x01
		xorwf	RECREG,w		; Test for GPO 1
		btfsc	STATUS, Z
		bcf	PORTB,GPO1		; Disable the GPO 1 output
		movlw	0x02
		xorwf	RECREG,w		; Test for GPO 2
		btfsc	STATUS, Z
		bcf	PORTB,GPO2		; Disable the GPO 2 output
		return

;-------------------------------------------------------------------------
; Set the contrast of LCD module
; Not avaiable for VFD
;-------------------------------------------------------------------------
SET_CONTRAST
		return

;-------------------------------------------------------------------------
; Set the Startup Screen
;-------------------------------------------------------------------------
SaveSplash
		movlw	.200
		call	XDELAY500		; delay 50ms, let command been transfered completely.

		movlw	low EE_SPLASH
		movwf	CONFIG_ADDR

		call	GetScreenSize
		movwf	TABLE_INDEX

SLOOP		call	GETCHAR
		movlw	MTRXCMD
		subwf	RECREG,w
		btfsc	STATUS,Z
		return				; rubish char received, leave the proc
		movf	RECREG,w
		movwf	CONFIG_DATA
		call	EE_WRITE
		incf	CONFIG_ADDR,f
		decfsz	TABLE_INDEX,f
		goto	SLOOP
		return

;-------------------------------------------------------------------------
; Calculate the screen size from screen configuration
;-------------------------------------------------------------------------
GetScreenSize
		clrf	DTMPW
		clrf	TABLE_INDEX
GLOOP		movf	DISPLAY_WIDTH,w
		addwf	DTMPW,f
		incf	TABLE_INDEX,f
		movf	TABLE_INDEX,w
		xorwf	DISPLAY_LINES,w
		btfss	STATUS,Z
		goto	GLOOP
		movf	DTMPW,w
		return

;-------------------------------------------------------------------------
; Move the LCD cursor to new position
;-------------------------------------------------------------------------
MOVE_CURSOR	movlw	0x80			; execute a cursor move command to D_ADDR
		iorwf	D_ADDR,w
		call	LCD_CMD
		return

;-------------------------------------------------------------------------
; Move the cursor to the right next position
;-------------------------------------------------------------------------
CUR_RIGHT
		incf	POS_X,f
		movf	DISPLAY_WIDTH,w
		xorwf	POS_X,w
		btfsc	STATUS,Z
		goto	GoNextLine		; current is end of line, should wrap to start of next line
		goto	CUR_END
GoNextLine	clrf	POS_X
		incf	POS_Y,f
		movf	DISPLAY_LINES,w
		xorwf	POS_Y,w
		btfsc	STATUS,Z		; check if new line is out of range?
		goto	WRAP_Line		; yes, it is last line, should wrap to first line or scroll
		goto	CUR_END			; no, just go ahead
Scroll_Line
		call	SCROLL_UP		; scroll the screen
		clrf	POS_X
		movlw	1
		subwf	DISPLAY_LINES,w
		movwf	POS_Y			; move to bottom line
		goto	CUR_END
WRAP_Line
		btfsc	FLAGS,AUTOSCROLL	; check the auto scroll flag?
		goto	Scroll_Line		; yes, should scroll the screen up
		clrf	POS_Y			; no, just return to the first line
CUR_END
		call	GET_ADDRESS
		call	MOVE_CURSOR
		return

;-------------------------------------------------------------------------
; Move the cursor to left position
;-------------------------------------------------------------------------
CUR_LEFT
		movf	POS_X,w
		btfsc	STATUS,Z
		goto	MoveUp
		decf	POS_X,f
		goto	CUR_L_END

MoveUp		decf	DISPLAY_WIDTH,w		; move cursor to end of upper line
		movwf	POS_X
		movf	POS_Y,w
		btfsc	STATUS,Z
		goto	TopLeft
		decf	POS_Y,f
		goto	CUR_L_END

TopLeft		clrf	POS_Y			; already at first line
		clrf	POS_X			; move cursor to start of first line

CUR_L_END
		call	GET_ADDRESS
		call	MOVE_CURSOR
		return

;-------------------------------------------------------------------------
; Transform the row/column value to DDRAM address
;-------------------------------------------------------------------------
GET_ADDRESS
		movf	POS_Y,w
		call	LINE_ADDRESS
		addwf	POS_X,w
		movwf	D_ADDR
		return

;-------------------------------------------------------------------------
; Lookup table contain DDRAM address of each line
;-------------------------------------------------------------------------
LINE_ADDRESS
		movwf	DTMPW
		movlw	HIGH DDRAM_ADDR
		movwf	PCLATH
		movf	DTMPW,w
		addwf	PCL ,F		;Jump to char pointed to in W reg
DDRAM_ADDR	dt	LCD_LINE0,LCD_LINE1,LCD_LINE2
DDRAM_ADDR_END	dt	LCD_LINE3
	IF ( (DDRAM_ADDR & 0x0FF) >= (DDRAM_ADDR_END & 0x0FF) )
		MESSG   "Warning - User Definded: Table 'DDRAM_ADDR' crosses page boundry in computed jump"
	ENDIF

;-------------------------------------------------------------------------
; Initialize the display setting
;-------------------------------------------------------------------------
Initialize
		movlw	low EE_TYPE		; init display type
		call	EE_READ
		movwf	DISPLAY_TYPE

		movlw	low EE_WIDTH		; init display width
		call	EE_READ
		andlw	0xff
		btfsc	STATUS,Z
		movlw	LCD_WIDTH		; default is 20
		movwf	DISPLAY_WIDTH

		movlw	low EE_LINES		; init display width
		call	EE_READ
		andlw	0xff
		btfsc	STATUS,Z
		movlw	LCD_LINES		; default is 2
		movwf	DISPLAY_LINES
		return

;-------------------------------------------------------------------------
; Initialize the horizontal bar chars
;-------------------------------------------------------------------------
InitHorizontalBar
		clrf	TABLE_INDEX
H_Start
		movf	TABLE_INDEX,w
		movwf	CG_ADDR
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f		; bits 3-5 now represent character number
		movf	CG_ADDR,w
		iorlw	b'01000000'		; load up a "set cgram address command"
		andlw	b'01111000'		; make sure only bits 3-6 are set at most
		call	LCD_CMD			; set CGRAM address

		movlw	8
		movwf	CurrentLine
		movf	TABLE_INDEX,w
		call	GetHorzByte
H1		call	LCD_CGPUTCHAR
		decfsz	CurrentLine,f
		goto	H1

		incf	TABLE_INDEX,f
		movlw	8
		xorwf	TABLE_INDEX,w
		btfss	STATUS,Z
		goto	H_Start

		movf	D_ADDR,w
		iorlw	b'10000000'		; set bit 7 to form the set ddram address cmd
		call	LCD_CMD			; set ddram address
		return

GetHorzByte
		movwf	DTMPW
		movlw	HIGH HorzByte
		movwf	PCLATH
		movf	DTMPW,w
		addwf	PCL ,F		;Jump to char pointed to in W reg
HorzByte	dt	b'10000',b'11000',b'11100',b'11110',b'11111',b'01111',b'00111'
HorzByte_End	dt	b'00011'
	IF ( (HorzByte & 0x0FF) >= (HorzByte_End & 0x0FF) )
		MESSG   "Warning - User Definded: Table 'HorzByte' crosses page boundry in computed jump"
	ENDIF
;-------------------------------------------------------------------------
; Initialize the Narrow Vertical bar chars
;-------------------------------------------------------------------------
InitNarrowVerticalBar
		movlw	b'00000011'
		movwf	G_CHAR
		goto	W_Start

;-------------------------------------------------------------------------
; Initialize the Wide Vertical bar chars
;-------------------------------------------------------------------------
InitWideVerticalBar
		movlw	b'00011111'
		movwf	G_CHAR

W_Start
		clrf	TABLE_INDEX

		movf	TABLE_INDEX,w
		andlw	b'00000111'		; only keep bits 0-2
		movwf	CG_ADDR
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f
		rlf	CG_ADDR,f		; bits 3-5 now represent character number
		movf	CG_ADDR,w
		iorlw	b'01000000'		; load up a "set cgram address command"
		andlw	b'01111000'		; make sure only bits 3-6 are set at most
		call	LCD_CMD			; set CGRAM address

W1		movlw	8
		movwf	CurrentLine
W2
		movf	CurrentLine,w
		movlw	1
		subwf	CurrentLine,w
		;addlw	1
		subwf	TABLE_INDEX,w
		btfsc	STATUS,C
		goto	W_LESS
		movlw	0
		goto	W_D
W_LESS		movf	G_CHAR,w
W_D
		call	LCD_CGPUTCHAR
		decfsz	CurrentLine,f
		goto	W2

		incf	TABLE_INDEX,f
		movlw	8
		xorwf	TABLE_INDEX,w
		btfss	STATUS,Z
		goto	W1
		movf	D_ADDR,w
		iorlw	b'10000000'		; set bit 7 to form the set ddram address cmd
		call	LCD_CMD			; set ddram address
		return

;-------------------------------------------------------------------------
; Places a bar graph at the specified column with the specified height.
;-------------------------------------------------------------------------
MakeVerticalBar
		call	GETCHAR			; get the column
		movf	RECREG,w
		movwf	POS_X
		call	GETCHAR			; get the height
		movf	RECREG,w
		movwf	CurrentLine

		movf	POS_X,w
		subwf	DISPLAY_WIDTH,w
		btfss	STATUS,C		; determine that if column is beyond the display width
		return				; yes, to large, so, ignore it
		movlw	1
		subwf	POS_X,f			; set the column position
		movf	DISPLAY_LINES,w
		movwf	POS_Y
V1
		decf	POS_Y,f			; set the line position
		call	GET_ADDRESS		; get the DDRAM address
		call	MOVE_CURSOR		; move cursor to the position

		movf	CurrentLine,w
		btfsc	STATUS,Z		; if height is 0
		goto	Zero			; then fill with space
		movlw	8
		subwf	CurrentLine,w		; calculate the last line hight
		btfss	STATUS,C
		goto	Less8			; height is less or equal to 8
Great8		movwf	CurrentLine		; height is great than 8
		movlw	7
		goto	Show

Less8
		andlw	0xff
		btfsc	STATUS,Z
		goto	Great8			; height is equal to 8
		movlw	1
		subwf	CurrentLine,w		; calculate the char for height
		movwf	DTMPW
		clrf	CurrentLine		; clear the height to 0 for the upcoming upper line
		movf	DTMPW,w
		goto	Show

Zero		movlw	0x20			; Space char
Show		call	LCD_PUTCHAR		; display the charactor
		movf	POS_Y,w
		andlw	0xff
		btfss	STATUS,Z
		goto	V1			; continue to process the upper line
		return

;-------------------------------------------------------------------------
; Places a horizontal bar graph at the specified column and row with the specified width.
; The format of the command is as follows:
;   <column number> 0x01 to 0x14 for a twenty column display
;   <row number> 0x01 to 0x02 for a two line display
;   <direction> 0 for left to right, 1 for right to left
;   <bar length> 0x00 to 0x64 (0 to100) for a twenty column display
;-------------------------------------------------------------------------
MakeHorizontalBar
		call	GETCHAR			; get the column
		movf	RECREG,w
		movwf	POS_X
		call	GETCHAR			; get the height
		movf	RECREG,w
		movwf	CurrentLine
		call	GETCHAR			; get the direction
		call	GETCHAR			; get bar length

		return

;-------------------------------------------------------------------------
; Show some debug information
;-------------------------------------------------------------------------
SHOW_INFO
		movlw	'W'
		call	LCD_PUTCHAR
		movlw	'='
		call	LCD_PUTCHAR
		movlw	low EE_WIDTH
		call	EE_READ
		xorlw	.20
		btfss	STATUS,Z
		goto	ERR1
		movlw	'2'
		call	LCD_PUTCHAR
		movlw	'0'
		call	LCD_PUTCHAR
		goto	STEP1
ERR1		movlw	'0'
		call	LCD_PUTCHAR
STEP1
		return

		include "usart.asm"	; The USART control routines are in file "usart.asm"
		include "lcdcmd.asm"	; The LCD control Routines are in File "lcdcmd.asm"

THE_END		goto THE_END

;-------------EEPROM----------------------------------------------
		org	H'2100'		; Initialize EEPROM Data
EE_WIDTH	de	D'20'		; The display width
EE_LINES	de	D'4'		; The lines
EE_TYPE		de	TYPE_LCD	; DISPLAY module flag
;EE_SPLASH		de	" Serial VFD Display "
;			de	" (c) 2002,Jerry Han "
EE_SPLASH	de	"+------------------+"
		de	"|Serial LCD Display|"
		de	"|(c)2002, Jerry Han|"
		de	"+------------------+"

;-------------the end----------------------------------------------
		end
